/* 
mIRC Classes by Yochai Timmer
yochai@dorot.org.il

Just thought of a way to make more efficient and complex data structures in mIRC
Hopefuly this will help people make some other interesting stuff for mIRC

v2.1
.Fixed a bug in /ClassMember not allowing $null values as default assignment
.Added $ObjectHash, returns the object's hash name, for fast hash table access
.Added scoping, to call methods from inherited classes
.Fixed wrong messasges in error handler
.Improved $object algorithm, added an exception to skip unnecessary parsing for simple cases
.Fixed a logical bug when using overloaded operators on objects (not members)
.Fixed a bug with conflicting regml results when using nested members repeatly
.Fixed a bug with passing paramters into For_Each_Class

v2.0
.Fixed a little logical bug in the classnew alias
.classread/classwrite
.Using regex
.fixed a logical bug in for_each_class when freeing
.added $this
.ClassMember and classinherit can only be called in constructor
.added a check in classfree to check if the class exists
.Changed the random number generator
.Members aren't intialized to 0 anymore, they can contain $null
.Changed function names


Thanks a lot for Dark_Rift and Mpdreamz for beta testing and giving me ideas and comments.
*/


/*
This function is used when you want to read or write data to\from your object.
When called as an identifier, this function will process its props as members, and function pointers.
When it's called as an alias it can be used as an assignment command (operators can be overloaded).
*/

alias Object {
  var %pointer = $1
  if ($hget(mirc_mem,$1)) var %type = $v1
  else return $ClassErrorHandle(Object,NO_SUCH_OBJECT,$1)
  var %class = $+(mem,$1)
  if ($isid) {
    var %members = $prop
    var %brace = $count(%members,$chr(40))
    if (%brace != $count(%members,$chr(41))) return $ClassErrorHandle(Object,WRONG_PARAMETERS,%members)
    if (!%members) return %type
    ;Fast result, no parsing
    if (%brace == 0) && (. !isin %members) {
      if (!$hfind(%class,%members,1)) return $ClassErrorHandle(Object,NO_SUCH_MEMBER,%members)
      return $hget(%class,%members)
    }
  }
  else {
    var %members = $2
    var %brace = $count(%members,$chr(40))
    if (%brace != $count(%members,$chr(41))) return $ClassErrorHandle(Object,WRONG_PARAMETERS,%members)
    var %operator = $3
    var %value = $4-
    ;check if $2 is the overloaded operator
    var %overload = $classGetMethod($+(operator,%members),%type,$hget(%class,classinherit))
    if ($isalias(%overload)) { 
      var %new_member
      var %value = $3-
      goto isid2
    }
    else if (%members == $chr(61)) { 
      var %new_member = $1
      var %value = $3-
      goto isid3
    }
    ;Fast result, no parsing
    if (%brace == 0) && (. !isin %members) {
      var %new_member = %members
      goto isid
    }
  }
  ;getting the pointer value, regex sperates members, methods
  var %re = /(?<=^|\.)([\x00-\x27\x2A-\x2D\x2F-\xFF]+(?:\((?:(?1)(?:\.(?1))*)?\))?|[^)]*)(?=\.|$)/g
  noop $regex(%members,%re)
  var %members
  var %a = 1
  while ($regml(%a)) { var %members = %members $v1 | inc %a  }
  var %a = 1
  while ($gettok(%members,%a,32)) {
    %new_member = $v1
    inc %a

    ;checking for valid pointer (to continue the class member parsing) 
    if ($hget(mirc_mem,%pointer)) var %type = $v1
    elseif (%pointer) return $ClassErrorHandle(Object,NO_SUCH_OBJECT,%pointer)
    else return $ClassErrorHandle(Object,NO_SUCH_MEMBER,%new_member)

    var %class = $+(mem,%pointer)
    %pointer = $hget(%class,%new_member)
    ;parsing command members
    if ($chr(40) isin %new_member) {
      var %inherit = $hget(%class,classinherit)
      var %method = $gettok(%new_member,1,40)
      if ($classGetMethod(%method,%type,%inherit)) var %function = $v1
      else {
        var %function = $hget(%class,%method)
        if (!$isalias(%function)) return $ClassErrorHandle(Object,NO_SUCH_METHOD,%method)
      }
      hadd -m mirc_memstack $calc($hget(mirc_memstack,0).item + 1) $right(%class,-3)
      %pointer = [ [ $+($,%function,$chr(40),$left($gettok(%new_member,2-,40),-1),$chr(41)) ] ]
      hdel mirc_memstack $hget(mirc_memstack,0).item
    }
    elseif (!$hfind(%class,%new_member,1)) return $ClassErrorHandle(Object,NO_SUCH_MEMBER,%new_member)
  }
  ;if the command was called as an assignment command
  if  (!$isid) {
    :isid
    if ($hget(mirc_mem,%new_member)) var %new_member = $null

    var %overload = $classGetMethod($+(operator,%operator),%type,$hget(%class,classinherit))
    if ($isalias(%overload)) {
      :isid2
      hadd -m mirc_memstack $calc($hget(mirc_memstack,0).item + 1) $right(%class,-3)
      [ [ $+($,%overload,$chr(40),%value,$chr(44),%new_member,$chr(41)) ] ]
      hdel mirc_memstack $hget(mirc_memstack,0).item
    }
    else if (%operator == $chr(61)) {
      :isid3
      if (%new_member) && ($hget(mirc_mem,%value)) classcopy %new_member %value
      else hadd %class %new_member %value
    } 
    else return $ClassErrorHandle(Object,NO_SUCH_OPERATOR,%operator)
    return
  }
  return %pointer
}

/*
This alias finds the apropriate method using the class unheritence list.
*/
alias -l classGetMethod {
  if (:: isin $1) {
    var %scope = $gettok($1,1,58)
    if (%scope != $2) && (!$istok($3,%scope,32) return $null
    else return $1
  }
  else var %function = $+($2,::,$1)
  if ($isalias(%function)) return %function
  var %inherit = $3
  var %a = 1
  while ($gettok(%inherit,%a,32)) {
    var %new = $v1
    var %function = $+(%new,::,$1)
    if ($isalias(%function)) return %function
    inc %a
  }
  return $null
}

/*
Returns calls the current class pointer (used inside methods and class functions)
*/
alias This {
  var %class = $hget(mirc_memstack,$hget(mirc_memstack,0).item)
  if (!$prop) return %class
  return $Object(%class) [ $+ . $+ [ $prop ]  ]
}

/* 
This function is a default copy function.
It just copies all the members from the right object to the left (this function is called by the class function when trying to copy an object to an object)
*/
alias -l ClassCopy {
  var %a = 1
  var %class1 = $+(mem,$1)
  var %class2 = $+(mem,$2)
  while ($hget(%class2,%a).item != $null) {
    inc %a
    var %member = $v1
    if (%member == classtype) || (%member == classinherit) continue
    if ($hget(%class1,%member) == $null) return $ClassErrorHandle(ClassCopy,CLASS_NO_MATCH) 
  }
  var %a = 1
  while ($hget(%class2,%a).item != $null) {
    inc %a
    var %member = $v1
    if (%member == classtype) || (%member == classinherit) continue
    hadd %class1 %member $hget(%class2,%member)
  }
}

/*
This function adds a member to a class.
This is the only valid way to add a member to a class, should usually be called by the constructor.
The function also allows you to add a starting value to the newly created member.
*/
alias ClassMember {
  var %class = $this
  if (!$hget(mirc_mem,%class)) return $ClassErrorHandle(ClassMember,NO_SUCH_OBJETC,%class)
  if ($hget($+(mem,%class),classtype)) return $ClassErrorHandle(ClassMember,NOT_IN_CONSTRUCTOR)
  if ($1 == classtype) || ($1 == classinherit) return $ClassErrorHandle(ClassMember,RESERVED_MEMBER,$1)
  if ($2) {
    if ($2 != $chr(61)) return $ClassErrorHandle(ClassMember,WRONG_PARAMETERS,$2-)
    hadd $+(mem,%class) $1 $3-
  }
  else {
    var %members = $1
    while ($gettok(%members,1,44)) { 
      var %members = $gettok(%members,2-,44)
      hadd $+(mem,%class) $1
    }
  }
}

/*
This function creates a new object by adding it to the objects database, and calling the class constructor if one exists.
The function will assign a pointer value to the class, and return it;
*/
alias NewObject {
  ; signature: $newobject([type])
  ; returns: pointer to newly created class
  var %check = $allocate_class($ticks $+ $hget(mirc_mem,0).item,$1)

  ;Checking if a class constructor exists
  if ($1) && ($isalias($+(class_,$1))) { 
    hadd -m mirc_memstack $calc($hget(mirc_memstack,0).item + 1) %check
    $+(class_,$1) $2-
    hdel mirc_memstack $hget(mirc_memstack,0).item
    hadd $+(mem,%check) classtype $1
  }
  return %check
}

alias -l allocate_class {
  var %point = $1
  hadd -m mirc_mem %point $iif($2,$2,1) 
  ;setting a default size of 10 (figured you don't usually need more, so lets save space)
  hmake $+(mem,%point) 10
  hadd -m $+(mirc_mem,$2) %point
  return %point
}

/*
This function will free the object pointed by the pointer.
It will call the class destructor if it exists and free the memory.
If the class inherits from a different class type, it will then call the inherited destructor.
*/
alias FreeObject {
  ;Checking if a class destructor exists
  var %class = $+(mem,$1)
  if (!$hget(mirc_mem,$1)) return $ClassErrorHandle(FreeObject,NO_SUCH_OBJECT,$1)
  var %type = $hget(%class,classtype)
  if (%type) && ($isalias($+(class~,%type))) {
    hadd -m mirc_memstack $calc($hget(mirc_memstack,0).item + 1) $1
    $+(class~,%type)
    hdel mirc_memstack $hget(mirc_memstack,0).item
  }
  var %inherit = $hget(%class,classinherit)
  var %destruct = $+(class~,$gettok(%inherit,1,32))
  if (%inherit) && ($isalias(%destruct)) {
    hadd %class classinherit $deltok(%inherit,1,32)
    hdel %class classtype
    hadd -m mirc_memstack $calc($hget(mirc_memstack,0).item + 1) $1
    %destruct %type
    hdel mirc_memstack $hget(mirc_memstack,0).item
  }
  if (!$2) || (%type == $2) {
    var %hashtype = $+(mirc_mem,%type)
    hdel %hashtype $1
    if ($hget(%hashtype,0).item == 0) hfree %hashtype 
    hdel mirc_mem $1
    hfree %class
  }
}

/*
This function makes the class inherit from a different class type.
This basicly means it will call the other type's constructor and run it on this class.
It will also remember the inherit type so it can call the inherited type's destructor when freed.
*/
alias ClassInherit {
  var %point = $this
  var %alias = $+(class_,$1)
  if ($hget($+(mem,%point),classtype)) return $ClassErrorHandle(ClassInherit,NOT_IN_CONSTRUCTOR)
  if ($1) && ($isalias(%alias)) {
    var %class = $+(mem,%point)
    var %inherits = $hget(%class,classinherit)
    if ($istok(%inherits,$1,32)) return ALREADY_INHERIT
    hadd -m mirc_memstack $calc($hget(mirc_memstack,0).item + 1) %point
    %alias $2-
    hdel mirc_memstack $hget(mirc_memstack,0).item
    hadd %class classinherit $addtok(%inherits,$1,32)
  }
  else return $ClassErrorHandle(ClassInherit,NO_SUCH_OBJECT)
}

/*
A loop that calls a function on each object of a certain class type, returns number of objects that were found
*/
alias For_Each_Class {
  var %hashtype = $+(mirc_mem,$1)
  var %a = 1
  if (!$2) return $ClassErrorHandle(For_Each_Class,WRONG_PARAMETERS,$1-)
  if (!$isalias($2)) return $ClassErrorHandle(For_Each_Class,NO_SUCH_ALIAS,$2)
  var %last
  while ($hget(%hashtype,%a).item) {
    var %pointer = $v1
    ;Safety to avoid user misuse
    if (%last == %pointer) inc %a
    var %last = %pointer
    hadd -m mirc_memstack $calc($hget(mirc_memstack,0).item + 1) %pointer
    if ( [ [ $+($,$2,$chr(40),$3-,$chr(41)) ] ] == 1) dec %a
    hdel mirc_memstack $hget(mirc_memstack,0).item
    inc %a
  }
  return $calc(%a -1)
}

/*
Returns the class type.
*/
alias ClassType {
  return $hget(mirc_mem,$1)
}

/*
Writes the class to an .ini file.
$SaveObject(pointer,file.ini,[section])
*/
alias SaveObject {
  if ($hget(mirc_mem,$1)) var %type = $v1
  else return $ClassErrorHandle(SaveObject,NO_SUCH_OBJECT,$1)
  if (!$isid) && (" isin $2) {
    var %file = $gettok($1-,2,34)
    var %section = $gettok($1-,3,34)
  }
  else {
    var %file = $noqt($2)
    var %section = $3
  }
  hsave -ai $+(mem,$1) $qt(%file) $iif(%section,%section,$+(%type,$chr(160),$1))
  return $null
}

/*
Loads a class from an .ini file, or an entire .ini file.
$LoadObject(file.ini,[section],[type])
*/
alias LoadObject {
  if (!$isid) && (" isin $1) {
    var %file = $gettok($1-,1,34)
    var %section = $gettok($gettok($1-,2-,34),1,32)
    var %type = $gettok($gettok($1-,2-,34),2,32)
  }
  else {
    var %file = $1
    var %section = $2
    var %type = $3
  }
  if (!$exists(%file)) return $ClassErrorHandle(LoadObject,NO_SUCH_FILE,%file)
  if (%section) {
    if ($ini(%file,%section)) return $_classread($noqt(%file),%section,%type)
    else return 0
  }
  else {
    var %a = 1
    var %count = 0
    while ($ini(%file,%a)) {
      var %temp = $_classread($noqt(%file), $v1, %type)
      if (%temp) inc %count
      inc %a
    }
    return $calc(%count)
  }
}

alias -l _classread {
  if (!$3) {
    var %check = $gettok($2,2,160)
    var %type = $gettok($2,1,160)
    ;try to use old pointer value
    if (%check !isnum) || ($hget(mirc_mem,%check)) var %point = $allocate_class($ticks $+ $hget(mirc_mem,0).item,%type)
    else var %point = $allocate_class(%check,%type)
  }
  else var %point = $NewObject($3)
  hload -i $+(mem,%point) $qt($1) $2
  return %point
}

/*
Returns the object's hash name
*/
alias ObjectHash {
  return $+(mem,$1)
}

/*
Error handler
*/
alias -l ClassErrorHandle {
  linesep
  var %error 
  if ($2 == NO_SUCH_OBJECT) var %error = Object $3 doesn't exist
  elseif ($2 == NO_SUCH_METHOD) var %error = Method $3 doesn't exist
  elseif ($2 == NO_SUCH_MEMBER) var %error = Member $3 doesn't exist
  elseif ($2 == NO_SUCH_OPERATOR) var %error = No operator overload for $3 
  elseif ($2 == CLASS_NO_MATCH) var %error = Class type mismatch
  elseif ($2 == WRONG_PARAMETERS) var %error = Bad parameters  $3 
  elseif ($2 == NO_SUCH_ALIAS) var %error = Alias  $3 wasn't found
  elseif ($2 == NOT_IN_CONSTRUCTOR) var %error = Can't use outside of constructor.
  elseif ($2 == RESERVED_MEMBER) var %error =   $3  is a saved member name.
  elseif ($2 == NO_SUCH_FILE) var %error = Can't find file: $3 
  echo $color(info) -s  * Error in $+(,$1,,:) %error
  linesep
  return $2
}
